import torch
import torch.nn as nn
import torchvision
from lightning import LightningDataModule
from lightning.pytorch import LightningModule
import torch.optim.optimizer
import wandb
from torchvision.transforms.v2 import ToPILImage
from nn_zoo.trainers.utils import get_optim, get_scheduler

__all__ = ["EncoderGANTrainer"]


class EncoderGANTrainer(LightningModule):
    def __init__(
        self,
        model: nn.Module,
        disc: nn.Module,
        dm: LightningDataModule,
        optim_g: str,
        optim_kwargs_g: dict,
        optim_d: str,
        optim_kwargs_d: dict,
        scheduler_g: str | None = None,
        scheduler_args_g: dict | None = None,
        scheduler_d: str | None = None,
        scheduler_args_d: dict | None = None,
    ):
        super(EncoderGANTrainer, self).__init__()
        self.model = model
        self.disc = disc
        self.dm = dm
        self.optim_g = get_optim(optim_g)
        self.optim_kwargs_g = optim_kwargs_g
        self.optim_d = get_optim(optim_d)
        self.optim_kwargs_d = optim_kwargs_d
        self.scheduler_g = get_scheduler(scheduler_g) if scheduler_g else None
        self.scheduler_kwargs_g = scheduler_args_g if scheduler_g else None
        self.scheduler_d = get_scheduler(scheduler_d) if scheduler_d else None
        self.scheduler_kwargs_d = scheduler_args_d if scheduler_d else None

        self.automatic_optimization = False

    def forward(self, x):
        return self.model(x).output

    def training_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        optim_g, optim_d = self.optimizers()

        # Train Discriminator
        optim_d.zero_grad()
        fake = self.model(x)
        real = x
        real_preds = self.disc(real)
        fake_preds = self.disc(fake)
        d_loss = self.disc.loss(real_preds, fake_preds)
        d_loss.backward()
        optim_d.step()

        # Train Generator
        optim_g.zero_grad()
        fake = self.model.generator(x)
        fake_preds = self.disc(fake)
        g_loss = self.model.generator_loss(fake_preds)
        g_loss.backward()
        optim_g.step()

        self.log("train/d_loss", d_loss)
        self.log("train/g_loss", g_loss)

        return d_loss + g_loss

        # preds = self.model(x)
        # metrics = self.model.loss(preds, x)

        # for key, value in metrics.items():
        #     self.log(f"train/{key}", value)

        # return metrics["loss"]

    def validation_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x)
        metrics = self.model.loss(preds, x)

        for key, value in metrics.items():
            self.log(f"val/{key}", value)

        if batch_idx == 0:
            self.logger.experiment.log(
                {
                    "val/imgs/recon_imgs": wandb.Image(
                        ToPILImage()(torchvision.utils.make_grid(preds).cpu()),
                        caption="Predictions",
                    ),
                }
            )

            if self.current_epoch == 0:
                self.logger.experiment.log(
                    {
                        "val/imgs/original_imgs": wandb.Image(
                            ToPILImage()(torchvision.utils.make_grid(x).cpu()),
                            caption="Original Images",
                        ),
                    }
                )

        return metrics["loss"]

    def test_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x)
        metrics = self.model.loss(preds, x)

        for key, value in metrics.items():
            self.log(f"test/{key}", value)

        if batch_idx == 0:
            self.logger.experiment.log(
                {
                    "test/imgs/recon_imgs": wandb.Image(
                        ToPILImage()(torchvision.utils.make_grid(preds).cpu()),
                        caption="Predictions",
                    ),
                }
            )

            self.logger.experiment.log(
                {
                    "test/imgs/negative": wandb.Image(
                        ToPILImage()(torchvision.utils.make_grid(x - preds).cpu()),
                        caption="Negative",
                    ),
                }
            )

        return metrics["loss"]

    def configure_optimizers(self):
        # optimizer = self.optim(self.model.parameters(), **self.optim_kwargs)
        # scheduler = (
        #     self.scheduler(optimizer, **self.scheduler_kwargs)  # type: ignore
        #     if self.scheduler
        #     else None
        # )

        # if scheduler:
        #     return [optimizer], [scheduler]
        # else:
        #     return optimizer

        optim_g = self.optim_g(self.model.generator.parameters(), **self.optim_kwargs_g)
        optim_d = self.optim_d(
            self.model.discriminator.parameters(), **self.optim_kwargs_d
        )
        scheduler_g = (
            self.scheduler_g(optim_g, **self.scheduler_kwargs_g)  # type: ignore
            if self.scheduler_g
            else None
        )
        scheduler_d = (
            self.scheduler_d(optim_d, **self.scheduler_kwargs_d)  # type: ignore
            if self.scheduler_d
            else None
        )

        if scheduler_g and scheduler_d:
            return [optim_g, optim_d], [scheduler_g, scheduler_d]
        elif scheduler_g:
            return [optim_g, optim_d], [scheduler_g]
        elif scheduler_d:
            return [optim_g, optim_d], [scheduler_d]
        else:
            return [optim_g, optim_d]

    def prepare_data(self) -> None:
        self.dm.prepare_data()

    def setup(self, stage):
        if stage == "fit":
            self.dm.setup("fit")
        elif stage == "test":
            self.dm.setup("test")

    def train_dataloader(self):
        return self.dm.train_dataloader()

    def val_dataloader(self):
        return self.dm.val_dataloader()

    def test_dataloader(self):
        return self.dm.test_dataloader()



import torch
from nn_zoo.models.components.depthwise import (
    DepthwiseConv1d,
    DepthwiseConv2d,
    DepthwiseConv3d,
    DepthwiseSeparableConv1d,
    DepthwiseSeparableConv2d,
    DepthwiseSeparableConv3d,
)


def test_depthwise_conv1d():
    B, T, C = 3, 32, 64
    x = torch.randn(B, C, T)

    conv = DepthwiseConv1d(C, 3)
    assert conv(x).shape == x.shape


def test_depthwise_conv2d():
    B, H, W, C = 3, 32, 32, 64
    x = torch.randn(B, C, H, W)

    conv = DepthwiseConv2d(C, 3)

    assert conv(x).shape == x.shape


def test_depthwise_conv3d():
    B, D, H, W, C = 3, 32, 32, 32, 64
    x = torch.randn(B, C, D, H, W)

    conv = DepthwiseConv3d(C, 3)

    assert conv(x).shape == x.shape


def test_depthwise_seperable_conv1d():
    B, T, C = 3, 32, 64
    x = torch.randn(B, C, T)

    conv = DepthwiseSeparableConv1d(C, C, 3)

    assert conv(x).shape == x.shape


def test_depthwise_seperable_conv2d():
    B, H, W, C = 3, 32, 32, 64
    x = torch.randn(B, C, H, W)

    conv = DepthwiseSeparableConv2d(C, C, 3)

    assert conv(x).shape == x.shape


def test_depthwise_seperable_conv3d():
    B, D, H, W, C = 3, 32, 32, 32, 64
    x = torch.randn(B, C, D, H, W)

    conv = DepthwiseSeparableConv3d(C, C, 3)

    assert conv(x).shape == x.shape



import torch


def get_optim(
    optim: str,
) -> type[torch.optim.Optimizer]:  # TODO: Add all the optimizer types
    """A function to get the optimizer class from the string
    Supports all the optimizers availible in PyTorch

    Args:
        optim (str): The optimizer name

    Raises:
        NotImplementedError: If the requested optimizer is not availible

    Returns:
        type[torch.optim.optimizer.Optimizer]: The optimizer class
    """
    match optim.lower():
        case "sgd":
            return torch.optim.SGD
        case "adam":
            return torch.optim.Adam
        case "adamw":
            return torch.optim.AdamW
        case _:
            raise NotImplementedError(
                f"The requested optimizer: {optim} is not availible"
            )


def get_scheduler(
    scheduler: str,
) -> type[torch.optim.lr_scheduler.LRScheduler]:  # TODO: Add all the scheduler types
    """A function to get the scheduler class from the string
    Supports all the schedulers availible in PyTorch

    Args:
        scheduler (str): The scheduler name

    Raises:
        NotImplementedError: If the requested scheduler is not availible

    Returns:
        type[torch.optim.lr_scheduler.LRScheduler]: The scheduler class
    """
    match scheduler.lower():
        case "steplr":
            return torch.optim.lr_scheduler.StepLR
        case "multisteplr":
            return torch.optim.lr_scheduler.MultiStepLR
        case "exponentiallr":
            return torch.optim.lr_scheduler.ExponentialLR
        case "cosinelr":
            return torch.optim.lr_scheduler.CosineAnnealingLR
        case _:
            raise NotImplementedError(
                f"The requested scheduler: {scheduler} is not availible"
            )



from typing import Callable

import torch
import torch.nn as nn
import torch.nn.functional as F
import pytest

from nn_zoo.models.utils import get_act, get_norm, RMSNorm1d, RMSNorm2d


@pytest.mark.parametrize(
    "act, expected",
    [
        ("relu", F.relu),
        ("leaky_relu", F.leaky_relu),
        ("tanh", F.tanh),
        ("sigmoid", F.sigmoid),
        ("prelu", F.prelu),
        ("silu", F.silu),
        ("gelu", F.gelu),
        ("softmax", F.softmax),
        ("identity", nn.Identity),
    ],
)
def test_get_act(act: str, expected: Callable[[torch.Tensor], torch.Tensor]):
    assert get_act(act) == expected  # type: ignore


@pytest.mark.parametrize(
    "norm, expected",
    [
        ("batch2d", nn.BatchNorm2d),
        ("instance2d", nn.InstanceNorm2d),
        ("rms1d", RMSNorm1d),
        ("rms2d", RMSNorm2d),
        ("layer", nn.LayerNorm),
        ("group", nn.GroupNorm),
        ("identity", nn.Identity),
    ],
)
def test_get_norm(norm: str, expected: Callable[[torch.Tensor], torch.Tensor]):
    assert get_norm(norm) == expected  # type: ignore


def test_rms1d_shape():
    ## 1D input
    x = torch.randn(3, 32)
    rms = RMSNorm1d(32)
    output = rms(x)

    assert output.shape == x.shape

    ## 2D input
    x = torch.randn(3, 32, 32)
    rms = RMSNorm1d(32)
    output = rms(x)

    assert output.shape == x.shape


def test_rms2d_shape():
    ## 2D input
    x = torch.randn(3, 32, 32)
    rms = RMSNorm2d(32)
    output = rms(x)

    assert output.shape == x.shape

    ## 3D input
    x = torch.randn(3, 32, 32, 32)
    rms = RMSNorm2d(32)
    output = rms(x)

    assert output.shape == x.shape



import torch
from nn_zoo.models.components.vq import VectorQuantizer


def test_vq():
    B, C, H, W = 3, 64, 32, 32
    x = torch.randn(B, C, H, W)

    vq = VectorQuantizer(C, 256, use_ema=True, decay=0.99, epsilon=1e-5)
    output = vq.forward(x)

    assert output[0].shape == (B, C, H, W)
    assert output[1] is None
    assert output[2].shape == ()  # Scalar tensor.
    assert output[3].shape == (B, H, W)



import torch
from torch import nn
import torch.nn.functional as F

__all__ = ["SelfAttention"]


class SelfAttention(nn.Module):
    def __init__(
        self, n_embd: int, n_head: int, attn_dropout: float = 0.0, is_causal=True
    ):
        super().__init__()
        assert n_embd % n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(n_embd, 3 * n_embd)
        # output projection
        self.c_proj = nn.Linear(n_embd, n_embd)
        # regularization
        self.n_head = n_head
        self.n_embd = n_embd
        self.attn_dropout = attn_dropout
        self.is_causal = is_causal

    def forward(self, x) -> torch.Tensor:
        B, T, C = x.size()

        qkv = self.c_attn(x)
        q, k, v = qkv.split(self.n_embd, dim=2)

        # (B, nh, T, hs)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)

        y = (
            F.scaled_dot_product_attention(
                q, k, v, is_causal=self.is_causal, dropout_p=self.attn_dropout
            )
            .transpose(1, 2)
            .contiguous()
            .view(B, T, C)
        )

        # output projection
        y = self.c_proj(y)
        return y



import torch
import torch.nn as nn

from lightning import LightningDataModule
from lightning.pytorch import LightningModule
from torchmetrics.functional.classification import accuracy

from nn_zoo.trainers.utils import get_optim, get_scheduler

__all__ = ["ClassifierTrainer"]


class ClassifierTrainer(LightningModule):
    def __init__(
        self,
        model: nn.Module,
        dm: LightningDataModule,
        optim: str,
        optim_kwargs: dict,
        scheduler: str | None = None,
        scheduler_args: dict | None = None,
    ):
        super(ClassifierTrainer, self).__init__()
        self.model = model
        self.dm = dm
        self.optim = get_optim(optim)
        self.optim_kwargs = optim_kwargs
        self.scheduler = get_scheduler(scheduler) if scheduler else None
        self.scheduler_kwargs = scheduler_args if scheduler else None

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def training_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x, y)
        loss = self.model.loss(preds, y)

        self.log("train_loss", loss)
        self.log(
            "train_acc",
            accuracy(
                preds.softmax(1),
                y,
                task="multiclass",
                num_classes=self.dm.num_classes(),
            ),
        )

        return loss

    def validation_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x, y)
        loss = self.model.loss(preds, y)

        self.log("val_loss", loss)
        self.log(
            "val_acc",
            accuracy(
                preds.softmax(1),
                y,
                task="multiclass",
                num_classes=self.dm.num_classes(),
            ),
        )

        return loss

    def test_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x, y)
        loss = self.model.loss(preds, y)

        self.log("test_loss", loss)
        self.log(
            "test_acc",
            accuracy(
                preds.softmax(1),
                y,
                task="multiclass",
                num_classes=self.dm.num_classes(),
            ),
        )

        return loss

    def configure_optimizers(self):
        optimizer = self.optim(self.model.parameters(), **self.optim_kwargs)
        scheduler = (
            self.scheduler(optimizer, **self.scheduler_kwargs)  # type: ignore
            if self.scheduler
            else None
        )

        if scheduler:
            return [optimizer], [scheduler]
        else:
            return optimizer

    def prepare_data(self) -> None:
        self.dm.prepare_data()

    def setup(self, stage):
        if stage == "fit":
            self.dm.setup("fit")
        elif stage == "test":
            self.dm.setup("test")

    def train_dataloader(self):
        return self.dm.train_dataloader()

    def val_dataloader(self):
        return self.dm.val_dataloader()

    def test_dataloader(self):
        return self.dm.test_dataloader()



import torch
from nn_zoo.models.components import SelfAttention


def test_causal_self_attention():
    B, T, C = 3, 32, 64
    x = torch.randn(B, T, C)

    attn = SelfAttention(C, 8)
    output = attn(x)

    assert output.shape == x.shape


def test_non_causal_self_attention():
    B, T, C = 3, 32, 64
    x = torch.randn(B, T, C)

    attn = SelfAttention(C, 8)
    output = attn(x)

    assert output.shape == x.shape



from torch.utils.data import DataLoader
from lightning import LightningDataModule
from torchvision import datasets, transforms


class FFHQDataModule(LightningDataModule):
    def __init__(
        self,
        data_dir: str,
        dataset_params: dict = {
            "download": True,
            "transform": transforms.ToTensor(),
        },
        loader_params: dict = {
            "batch_size": 64,
            "num_workers": 2,
        },
        thumbnails: bool = True,
    ):
        super().__init__()
        self.data_dir = data_dir
        self.dataset_params = dataset_params
        self.loader_params = loader_params
        self.thumbnails = thumbnails
        self.source = datasets.ImageFolder

        self.transforms = transforms.Compose([self.dataset_params["transform"]])
        self.dataset_params.__delitem__("transform")

    def __repr__(self) -> str:
        return f"{self.source.__name__}DataModule({self.data_dir})"

    def config(self) -> dict:
        return dict(
            data_dir=self.data_dir,
            dataset_params=self.dataset_params,
            loader_params=self.loader_params,
            thumbnails=self.thumbnails,
        )

    def prepare_data(self):
        if self.thumbnails:
            raise NotImplementedError("Thumbnails downloads not supported yet")

        else:
            raise NotImplementedError("Full dataset downloads not supported yet")

    def setup(self, stage=None):
        self.train_dataset = self.source(
            self.data_dir, **self.dataset_params, transform=self.transforms
        )
        self.val_dataset = self.source(
            self.data_dir, **self.dataset_params, transform=self.transforms
        )
        self.test_dataset = self.source(
            self.data_dir, **self.dataset_params, transform=self.transforms
        )

    def train_dataloader(self):
        assert self.train_dataset is not None, f"{self.__class__} not setup properly"
        return DataLoader(self.train_dataset, shuffle=True, **self.loader_params)

    def val_dataloader(self):
        assert self.val_dataset is not None, f"{self.__class__} not setup properly"
        return DataLoader(self.val_dataset, **self.loader_params)

    def test_dataloader(self):
        assert self.test_dataset is not None, f"{self.__class__} not setup properly"
        return DataLoader(self.test_dataset, **self.loader_params)

    def num_classes(self):
        return 0


if __name__ == "__main__":
    data_module = FFHQDataModule("data")
    data_module.prepare_data()
    data_module.setup()
    train = data_module.train_dataloader()
    val = data_module.val_dataloader()
    test = data_module.test_dataloader()
    print(f"Train: {len(data_module.train_dataset):,}")
    print(f"Val: {len(data_module.val_dataset):,}")
    print(f"Test: {len(data_module.test_dataset):,}")
    print(data_module)



from .attention import SelfAttention
from .residual import ResidualBasicBlock, ResidualBottleNeckBlock, ResidualStack
from .vq import VectorQuantizer
from .depthwise import (
    DepthwiseConv1d,
    DepthwiseConv2d,
    DepthwiseConv3d,
    DepthwiseSeparableConv1d,
    DepthwiseSeparableConv2d,
    DepthwiseSeparableConv3d,
)

__all__ = [
    "DepthwiseConv1d",
    "DepthwiseConv2d",
    "DepthwiseConv3d",
    "DepthwiseSeparableConv1d",
    "DepthwiseSeparableConv2d",
    "DepthwiseSeparableConv3d",
    "PCA",
    "SelfAttention",
    "ResidualBasicBlock",
    "ResidualBottleNeckBlock",
    "ResidualStack",
    "VectorQuantizer",
]



import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Literal
from collections.abc import Callable

__all__ = ["get_act", "get_norm", "RMSNorm1d", "RMSNorm2d"]


def get_act(
    act: Literal[
        "relu",
        "leaky_relu",
        "tanh",
        "sigmoid",
        "prelu",
        "silu",
        "gelu",
        "softmax",
        "identity",
    ]
    | Callable[[torch.Tensor], torch.Tensor],
) -> Callable[[torch.Tensor], torch.Tensor]:
    """Return the activation function based on the string name.

    You can use this function to get the activation function based on the string name or pass in a desired function directly.

    Args:
        act (str): The name of the activation function.

    Returns:
        Callable[[torch.Tensor], torch.Tensor]: The activation function.
    """
    if callable(act):
        return act

    return {
        "relu": F.relu,
        "leaky_relu": F.leaky_relu,
        "tanh": F.tanh,
        "sigmoid": F.sigmoid,
        "prelu": F.prelu,
        "silu": F.silu,
        "gelu": F.gelu,
        "softmax": F.softmax,
        "identity": nn.Identity,
    }[act]


def get_norm(
    norm: Literal[
        "batch2d",
        "instance2d",
        "rms1d",
        "rms2d",
        "layer",
        "group",
        "identity",
    ],
) -> type[nn.Module]:
    """Return the normalization layer based on the string name.

    You can use this function to get the normalization layer based on the string name.

    Args:
        norm (str): The name of the normalization layer.

    Returns:
        type[nn.Module]: The normalization layer.
    """

    return {
        "batch2d": nn.BatchNorm2d,
        "instance2d": nn.InstanceNorm2d,
        "rms1d": RMSNorm1d,
        "rms2d": RMSNorm2d,
        "layer": nn.LayerNorm,
        "group": nn.GroupNorm,
        "identity": nn.Identity,
    }[norm]


class RMSNorm1d(nn.Module):
    """
    Implements Root Mean Square Normalization introduced in
    https://arxiv.org/pdf/1910.07467.pdf.

    Reference implementation (used for correctness verfication)
    can be found here:
    https://github.com/facebookresearch/llama/blob/main/llama/model.py

    Args:
        dim (int): embedding size
        eps (float): small value to avoid division by zero. Default: 1e-6
    """

    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x (torch.Tensor): input tensor to normalize

        Returns:
            torch.Tensor: The output tensor after applying RMSNorm.
        """
        # computation is in fp32
        x_fp32 = x.float()
        x_normed = (
            x_fp32 * torch.rsqrt(x_fp32.pow(2).mean(-1, keepdim=True) + self.eps)
        ).type_as(x)
        return x_normed * self.scale


class RMSNorm2d(nn.Module):
    """
    Implements Root Mean Square Normalization introduced in
    https://arxiv.org/pdf/1910.07467.pdf.

    Args:
        dim (int): embedding size
        eps (float): small value to avoid division by zero. Default: 1e-6
    """

    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x (torch.Tensor): input tensor to normalize

        Returns:
            torch.Tensor: The output tensor after applying RMSNorm.
        """
        # computation is in fp32
        x_fp32 = x.float()
        x_normed = (
            x_fp32 * torch.rsqrt(x_fp32.pow(2).mean(-1, keepdim=True) + self.eps)
        ).type_as(x)
        return x_normed * self.scale



"""Depthwise Convolution Modules from https://arxiv.org/abs/1610.02357"""

from torch import nn

__all__ = [
    "DepthwiseConv1d",
    "DepthwiseConv2d",
    "DepthwiseConv3d",
    "DepthwiseSeparableConv1d",
    "DepthwiseSeparableConv2d",
    "DepthwiseSeparableConv3d",
]


class DepthwiseConv1d(nn.Conv1d):
    """Depthwise Convolution for 1D data."""

    def __init__(
        self,
        in_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
    ):
        super(DepthwiseConv1d, self).__init__(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )


class DepthwiseConv2d(nn.Conv2d):
    """Depthwise Convolution for 2D data."""

    def __init__(
        self,
        in_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
    ):
        super(DepthwiseConv2d, self).__init__(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )


class DepthwiseConv3d(nn.Conv3d):
    """Depthwise Convolution for 3D data."""

    def __init__(
        self,
        in_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
    ):
        super(DepthwiseConv3d, self).__init__(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )


class DepthwiseSeparableConv1d(nn.Sequential):
    """Depthwise Separable Convolution for 1D data."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        bias: bool = True,
    ):
        super(DepthwiseSeparableConv1d, self).__init__(
            DepthwiseConv1d(
                in_channels,
                kernel_size,
                stride,
                padding,
                dilation,
                groups=in_channels,
                bias=bias,
            ),
            nn.Conv1d(in_channels, out_channels, 1, bias=bias),
        )


class DepthwiseSeparableConv2d(nn.Sequential):
    """Depthwise Separable Convolution for 2D data."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        bias: bool = True,
    ):
        super(DepthwiseSeparableConv2d, self).__init__(
            DepthwiseConv2d(
                in_channels,
                kernel_size,
                stride,
                padding,
                dilation,
                groups=in_channels,
                bias=bias,
            ),
            nn.Conv2d(in_channels, out_channels, 1, bias=bias),
        )


class DepthwiseSeparableConv3d(nn.Sequential):
    """Depthwise Separable Convolution for 3D data."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        bias: bool = True,
    ):
        super(DepthwiseSeparableConv3d, self).__init__(
            DepthwiseConv3d(
                in_channels,
                kernel_size,
                stride,
                padding,
                dilation,
                groups=in_channels,
                bias=bias,
            ),
            nn.Conv3d(in_channels, out_channels, 1, bias=bias),
        )



import torch
from nn_zoo.models.components.residual import (
    ResidualBasicBlock,
    ResidualBottleNeckBlock,
    ResidualStack,
)


def test_basic_residual_block():
    B, C, H, W = 3, 32, 32, 32
    x = torch.randn(B, C, H, W)

    block = ResidualBasicBlock(C, C, 3)
    assert block(x).shape == x.shape


def test_bottleneck_residual_block():
    B, C, H, W = 3, 32, 32, 32
    x = torch.randn(B, C, H, W)

    block = ResidualBottleNeckBlock(C, C, 3)
    assert block(x).shape == x.shape


def test_residual_stack():
    B, C, H, W = 3, 32, 32, 32
    x = torch.randn(B, C, H, W)

    stack = ResidualStack(3, C, C, 3)
    assert stack(x).shape == x.shape



from typing import Callable
import torch
from torch import nn
from torch.nn import functional as F

__all__ = ["ResidualBasicBlock", "ResidualBottleNeckBlock", "ResidualStack"]


class ResidualBasicBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        act: Callable[[torch.Tensor], torch.Tensor] = F.relu,
        norm_layer: type[nn.Module] | None = nn.BatchNorm2d,
    ):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size, stride, padding)

        self.norm1 = (
            norm_layer(out_channels) if norm_layer is not None else nn.Identity()
        )
        self.norm2 = (
            norm_layer(out_channels) if norm_layer is not None else nn.Identity()
        )
        self.act = act

        self.skip = (
            nn.Conv2d(in_channels, out_channels, kernel_size=1)
            if in_channels != out_channels
            else nn.Identity()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.act(x)
        x = self.conv2(x)
        x = self.norm2(x)
        x = x + self.skip(residual)
        x = self.act(x)
        return x


class ResidualBottleNeckBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        act: Callable[[torch.Tensor], torch.Tensor] = F.relu,
        norm_layer: type[nn.Module] | None = nn.BatchNorm2d,
    ):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size, stride, padding)
        self.conv3 = nn.Conv2d(out_channels, out_channels, kernel_size, stride, padding)

        self.norm1 = (
            norm_layer(out_channels) if norm_layer is not None else nn.Identity()
        )
        self.norm2 = (
            norm_layer(out_channels) if norm_layer is not None else nn.Identity()
        )
        self.norm3 = (
            norm_layer(out_channels) if norm_layer is not None else nn.Identity()
        )
        self.act = act

        self.skip = (
            nn.Conv2d(in_channels, out_channels, 1)
            if in_channels != out_channels
            else nn.Identity()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.act(x)
        x = self.conv2(x)
        x = self.norm2(x)
        x = self.act(x)
        x = self.conv3(x)
        x = self.norm3(x)
        x += self.skip(residual)
        x = self.act(x)
        return x


class ResidualStack(nn.Sequential):
    def __init__(
        self,
        n_blocks: int,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        act: Callable[[torch.Tensor], torch.Tensor] = F.relu,
        norm_layer: type[nn.Module] | None = nn.BatchNorm2d,
        block: type[nn.Module] = ResidualBasicBlock,
    ):
        super().__init__(
            *[
                block(
                    in_channels if i == 0 else out_channels,
                    out_channels,
                    kernel_size,
                    stride,
                    padding,
                    act,
                    norm_layer,
                )
                for i in range(n_blocks)
            ]
        )



import torch
import torch.nn as nn
import torchvision
from lightning import LightningDataModule
from lightning.pytorch import LightningModule
import torch.optim.optimizer
import wandb
from torchvision.transforms.v2 import ToPILImage
from nn_zoo.trainers.utils import get_optim, get_scheduler

__all__ = ["AutoEncoderTrainer"]


class AutoEncoderTrainer(LightningModule):
    def __init__(
        self,
        model: nn.Module,
        dm: LightningDataModule,
        optim: str,
        optim_kwargs: dict,
        scheduler: str | None = None,
        scheduler_args: dict | None = None,
    ):
        super(AutoEncoderTrainer, self).__init__()
        self.model = model
        self.dm = dm
        self.optim = get_optim(optim)
        self.optim_kwargs = optim_kwargs
        self.scheduler = get_scheduler(scheduler) if scheduler else None
        self.scheduler_kwargs = scheduler_args if scheduler else None

    def forward(self, x):
        return self.model(x).output

    def training_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x)
        metrics = self.model.loss(preds, x)

        for key, value in metrics.items():
            self.log(f"train/{key}", value)

        return metrics["loss"]

    def validation_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x)
        metrics = self.model.loss(preds, x)

        for key, value in metrics.items():
            self.log(f"val/{key}", value)

        if batch_idx == 0:
            self.logger.experiment.log(
                {
                    "val/imgs/recon_imgs": wandb.Image(
                        ToPILImage()(torchvision.utils.make_grid(preds).cpu()),
                        caption="Predictions",
                    ),
                }
            )

            if self.current_epoch == 0:
                self.logger.experiment.log(
                    {
                        "val/imgs/original_imgs": wandb.Image(
                            ToPILImage()(torchvision.utils.make_grid(x).cpu()),
                            caption="Original Images",
                        ),
                    }
                )

        return metrics["loss"]

    def test_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        x, y = batch

        preds = self.model(x)
        metrics = self.model.loss(preds, x)

        for key, value in metrics.items():
            self.log(f"test/{key}", value)

        if batch_idx == 0:
            self.logger.experiment.log(
                {
                    "test/imgs/recon_imgs": wandb.Image(
                        ToPILImage()(torchvision.utils.make_grid(preds).cpu()),
                        caption="Predictions",
                    ),
                }
            )

            self.logger.experiment.log(
                {
                    "test/imgs/negative": wandb.Image(
                        ToPILImage()(torchvision.utils.make_grid(x - preds).cpu()),
                        caption="Negative",
                    ),
                }
            )

        return metrics["loss"]

    def configure_optimizers(self):
        optimizer = self.optim(self.model.parameters(), **self.optim_kwargs)
        scheduler = (
            self.scheduler(optimizer, **self.scheduler_kwargs)  # type: ignore
            if self.scheduler
            else None
        )

        if scheduler:
            return [optimizer], [scheduler]
        else:
            return optimizer

    def prepare_data(self) -> None:
        self.dm.prepare_data()

    def setup(self, stage):
        if stage == "fit":
            self.dm.setup("fit")
        elif stage == "test":
            self.dm.setup("test")

    def train_dataloader(self):
        return self.dm.train_dataloader()

    def val_dataloader(self):
        return self.dm.val_dataloader()

    def test_dataloader(self):
        return self.dm.test_dataloader()



import torch
from torch import nn
import torch.nn.functional as F

__all__ = ["VectorQuantizer"]


class SonnetExponentialMovingAverage(nn.Module):
    # See: https://github.com/deepmind/sonnet/blob/5cbfdc356962d9b6198d5b63f0826a80acfdf35b/sonnet/src/moving_averages.py#L25.
    # They do *not* use the exponential moving average updates described in Appendix A.1
    # of "Neural Discrete Representation Learning".
    def __init__(self, decay, shape):
        super().__init__()
        self.decay = decay
        self.counter = 0
        self.register_buffer("hidden", torch.zeros(*shape))
        self.register_buffer("average", torch.zeros(*shape))

    def update(self, value):
        self.counter += 1
        with torch.no_grad():
            self.hidden -= (self.hidden - value) * (1 - self.decay)
            self.average = self.hidden / (1 - self.decay**self.counter)

    def __call__(self, value):
        self.update(value)
        return self.average


class VectorQuantizer(nn.Module):
    def __init__(self, embedding_dim, num_embeddings, use_ema, decay, epsilon):
        super().__init__()
        # See Section 3 of "Neural Discrete Representation Learning" and:
        # https://github.com/deepmind/sonnet/blob/v2/sonnet/src/nets/vqvae.py#L142.

        self.embedding_dim = embedding_dim
        self.num_embeddings = num_embeddings
        self.use_ema = use_ema
        # Weight for the exponential moving average.
        self.decay = decay
        # Small constant to avoid numerical instability in embedding updates.
        self.epsilon = epsilon

        # Dictionary embeddings.
        limit = 3**0.5
        e_i_ts = torch.FloatTensor(embedding_dim, num_embeddings).uniform_(
            -limit, limit
        )
        if use_ema:
            self.register_buffer("e_i_ts", e_i_ts)
        else:
            self.register_parameter("e_i_ts", nn.Parameter(e_i_ts))

        # Exponential moving average of the cluster counts.
        self.N_i_ts = SonnetExponentialMovingAverage(decay, (num_embeddings,))
        # Exponential moving average of the embeddings.
        self.m_i_ts = SonnetExponentialMovingAverage(decay, e_i_ts.shape)

    def forward(self, x):
        flat_x = x.permute(0, 2, 3, 1).reshape(-1, self.embedding_dim)
        distances = (
            (flat_x**2).sum(1, keepdim=True)
            - 2 * flat_x @ self.e_i_ts
            + (self.e_i_ts**2).sum(0, keepdim=True)
        )
        encoding_indices = distances.argmin(1)
        quantized_x = F.embedding(
            encoding_indices.view(x.shape[0], *x.shape[2:]), self.e_i_ts.transpose(0, 1)
        ).permute(0, 3, 1, 2)

        # See second term of Equation (3).
        if not self.use_ema:
            dictionary_loss = ((x.detach() - quantized_x) ** 2).mean()
        else:
            dictionary_loss = None

        # See third term of Equation (3).
        commitment_loss = ((x - quantized_x.detach()) ** 2).mean()
        # Straight-through gradient. See Section 3.2.
        quantized_x = x + (quantized_x - x).detach()

        if self.use_ema and self.training:
            with torch.no_grad():
                # See Appendix A.1 of "Neural Discrete Representation Learning".

                # Cluster counts.
                encoding_one_hots = F.one_hot(
                    encoding_indices, self.num_embeddings
                ).type(flat_x.dtype)
                n_i_ts = encoding_one_hots.sum(0)
                # Updated exponential moving average of the cluster counts.
                # See Equation (6).
                self.N_i_ts(n_i_ts)

                # Exponential moving average of the embeddings. See Equation (7).
                embed_sums = flat_x.transpose(0, 1) @ encoding_one_hots
                self.m_i_ts(embed_sums)

                # This is kind of weird.
                # Compare: https://github.com/deepmind/sonnet/blob/v2/sonnet/src/nets/vqvae.py#L270
                # and Equation (8).
                N_i_ts_sum = self.N_i_ts.average.sum()
                N_i_ts_stable = (
                    (self.N_i_ts.average + self.epsilon)
                    / (N_i_ts_sum + self.num_embeddings * self.epsilon)
                    * N_i_ts_sum
                )
                self.e_i_ts = self.m_i_ts.average / N_i_ts_stable.unsqueeze(0)

        return (
            quantized_x,
            dictionary_loss,
            commitment_loss,
            encoding_indices.view(x.shape[0], *x.shape[2:]),
        )
